#include "uploader.h"
#include <cstring>

Uploader::Uploader(QObject *parent) : QObject(parent), m_context(this), m_allocated(0), m_program(this), m_jpegs(0)
{
    // "applications must ensure that create() is only called on the main (GUI) thread"
    m_surface.create();

    m_textures = QSharedPointer<QVector<GLuint>>(new QVector<GLuint>);

    // reserve space for 10 rectangles
    m_data.reserve(280);
    m_textures->reserve(10);
}

Uploader::~Uploader()
{
    m_program.release();
}

void Uploader::createContext(QOpenGLContext *context)
{
    static const float position[] = {-1.0f, -1.0f,
                                     -1.0f, 1.0f,
                                     1.0f, -1.0f,
                                     1.0f, 1.0f};
  
    m_context.setShareContext(context);
    m_context.create();
    m_context.makeCurrent(&m_surface);
    m_functions = m_context.functions();

    m_program.addShaderFromSourceFile(QOpenGLShader::Vertex, "copy_shader.vsh");
    m_program.addShaderFromSourceFile(QOpenGLShader::Fragment, "copy_shader.fsh");
    m_program.bind();
    m_functions->glActiveTexture(GL_TEXTURE0);
    m_program.setUniformValue("texture", 0);
    m_program.enableAttributeArray("position");
    m_program.enableAttributeArray("texcoord");
    m_program.setAttributeArray("position", GL_FLOAT, position, 2);

    m_vertices.create();
    m_vertices.setUsagePattern(QOpenGLBuffer::DynamicDraw);
}

void Uploader::cleanup()
{
    m_data.clear();
    m_functions->glDeleteTextures(m_textures->size(), m_textures->constData());
    m_textures->clear();
    m_finished = false;
}

void Uploader::gotFill(const int x, const int y, const int width, const int height, const float red, const float green, const float blue)
{
    m_textures->append(0);
    // NOTE: -coordinates are transformed to normalized device coordinates in the vertex shader
    //       -coordinates and colors are extended to vec4 in the vertex shader
    //       -OpenGL ES 2.0 does not support integer values as input,
    //        thus we need to cast to float on the CPU side :(
    // bottom left
    m_data.append((float) x);
    m_data.append((float) (y + height));
    m_data.append(red);
    m_data.append(green);
    m_data.append(blue);
    m_data.append(0.0f);
    m_data.append(0.0f);
    // top left
    m_data.append((float) x);
    m_data.append((float) y);
    m_data.append(red);
    m_data.append(green);
    m_data.append(blue);
    m_data.append(0.0f);
    m_data.append(0.0f);
    // bottom right
    m_data.append((float) (x + width));
    m_data.append((float) (y + height));
    m_data.append(red);
    m_data.append(green);
    m_data.append(blue);
    m_data.append(0.0f);
    m_data.append(0.0f);
    // top right
    m_data.append((float) (x + width));
    m_data.append((float) y);
    m_data.append(red);
    m_data.append(green);
    m_data.append(blue);
    m_data.append(0.0f);
    m_data.append(0.0f);
}

void Uploader::gotCopy(const int src_x, const int src_y, const int width, const int height, const int dest_x, const int dest_y)
{
    float texcoord[] = {(float) src_x,              (float) (src_y + height),
                        (float) src_x,              (float) src_y,
                        (float) (src_x + width),    (float) (src_y + height),
                        (float) (src_x + width),    (float) src_y};

    QOpenGLFramebufferObject fbo(width, height);
    fbo.bind();
    m_functions->glViewport(0, 0, width, height);
    m_program.setAttributeArray("texcoord", GL_FLOAT, texcoord, 2);
    m_functions->glBindTexture(GL_TEXTURE_2D, m_FBOTexture);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    m_functions->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    m_functions->glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    m_textures->append(fbo.takeTexture());
    
    // bottom left
    m_data.append((float) dest_x);
    m_data.append((float) (dest_y + height));

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(0.0f);
    m_data.append(0.0f);
    // top left
    m_data.append((float) dest_x);
    m_data.append((float) dest_y);

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(0.0f);
    m_data.append(1.0f);
    // bottom right
    m_data.append((float) (dest_x + width));
    m_data.append((float) (dest_y + height));

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(1.0f);
    m_data.append(0.0f);
    // top right
    m_data.append((float) (dest_x + width));
    m_data.append((float) dest_y);

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(1.0f);
    m_data.append(1.0f);
}

void Uploader::gotBitmap(const QByteArray &bitmap, const quint32 x, const quint32 y,
                         const quint32 width, const quint32 height, const GLenum format, const GLenum type)
{
    quint32 real_width = qNextPowerOfTwo(width - 1);
    quint32 real_height = qNextPowerOfTwo(height - 1);

    GLuint id;
    m_functions->glGenTextures(1, &id);
    m_functions->glBindTexture(GL_TEXTURE_2D, id);
    m_functions->glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, real_width, real_height, 0, format, type, NULL);
    m_functions->glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, format, type, bitmap.constData());
    m_textures->append(id);
    float bottom = (float) height / (float) real_height;
    float right = (float) width / (float) real_width;

    // bottom left
    m_data.append((float) x);
    m_data.append((float) (y + height));

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(0.0f);
    m_data.append(bottom);
    // top left
    m_data.append((float) x);
    m_data.append((float) y);

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(0.0f);
    m_data.append(0.0f);
    // bottom right
    m_data.append((float) (x + width));
    m_data.append((float) (y + height));

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(right);
    m_data.append(bottom);
    // top right
    m_data.append((float) (x + width));
    m_data.append((float) y);

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(right);
    m_data.append(0.0f);
}

// Allocate storage for the image
int Uploader::gotJpeg(const quint32 x, const quint32 y, const quint32 width, const quint32 height)
{
    m_jpegs++;

    // "This function returns the nearest power of two greater than value."
    // We need greater than or equal
    quint32 real_width = qNextPowerOfTwo(width - 1);
    quint32 real_height = qNextPowerOfTwo(height - 1);

    GLuint id;
    m_functions->glGenTextures(1, &id);
    m_functions->glBindTexture(GL_TEXTURE_2D, id);
#ifdef GL_RGB8
    m_functions->glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, real_width, real_height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
#else
    #ifdef GL_RGB8_OES
        m_functions->glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8_OES, real_width, real_height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
    #else
        m_functions->glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, real_width, real_height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
    #endif
#endif

    m_textures->append(id);

    float bottom = (float) height / (float) real_height;
    float right = (float) width / (float) real_width;

    // bottom left
    m_data.append((float) x);
    m_data.append((float) (y + height));

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(0.0f);
    m_data.append(bottom);
    // top left
    m_data.append((float) x);
    m_data.append((float) y);

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(0.0f);
    m_data.append(0.0f);
    // bottom right
    m_data.append((float) (x + width));
    m_data.append((float) (y + height));

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(right);
    m_data.append(bottom);
    // top right
    m_data.append((float) (x + width));
    m_data.append((float) y);

    m_data.append(0.0f);
    m_data.append(0.0f);
    m_data.append(0.0f);

    m_data.append(right);
    m_data.append(0.0f);

    return m_textures->size() - 1;
}

void Uploader::finishedJpeg(const unsigned char *image, const quint32 width, const quint32 height, const int index)
{
    m_jpegs--;
    m_functions->glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    m_functions->glBindTexture(GL_TEXTURE_2D, m_textures->at(index));
    m_functions->glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, image);
    delete[] image;
    if(m_jpegs == 0 && m_finished) {
        m_functions->glBindTexture(GL_TEXTURE_2D, 0);
        m_functions->glFinish();
        emit uploadFinished();
    }
}

void Uploader::finishedUpdate()
{
    unsigned int size = m_data.size() * sizeof(float);
    unsigned int nextPowerOfTwo = qNextPowerOfTwo(size);

    m_vertices.bind();
    
    if(nextPowerOfTwo > m_allocated) {
      m_vertices.allocate(nextPowerOfTwo);
      m_allocated = nextPowerOfTwo;
    }

    m_vertices.write(0, m_data.constData(), size);
    
    m_vertices.release();

    if(m_jpegs == 0) {
        m_functions->glBindTexture(GL_TEXTURE_2D, 0);
        emit uploadFinished();
    } else {
        m_finished = true;
    }
}

void Uploader::swapBuffers()
{
    // This thread should finish uploading
    // The other thread should finish rendering from the other thread
    m_functions->glFinish();
}

QSharedPointer<QVector<GLuint>> Uploader::textures() const
{
    return m_textures;
}

QOpenGLBuffer Uploader::vertices() const
{
  return m_vertices;
}

void Uploader::changeFBOTexture(const unsigned int FBOTexture, const QMatrix4x4 &ortho)
{
    m_FBOTexture = FBOTexture;
    m_program.setUniformValue("ortho", ortho);
}
